from tkinter import *
import tkinter.font
import json
from re import match
from types import NoneType
from cytomat_plus.db import DB
#from cytomat_plus.scripts.setup_cytomat_plus import get_config_dir

class ManagerConfigFile():
    __root: Frame
    __db: DB
    def __init__(self, root, config_directory_path, data, db: DB) -> None:
        self.__root = root
        self.__data = data
        self.__db = db
        self.__config_directory_path = config_directory_path
        self.create_overview()
        
    def create_overview(self):
        self.set_grid_settings()
        self.create_header_config()
        self.create_warning_message()
        self.create_row_obejcts()
        
    def create_header_config(self):
        self.__header_config = Label(self.__root, text = "Set specific parameters for the Cytomat", font=('Helvetica', 20), pady = 1, padx = 1, relief='solid')
        self.__header_config.grid(row = 0, column = 0, sticky = 'nsew', columnspan = 3, ipady=30)
    
    def create_warning_message(self):
        warning_msg =  """Warning Message: blablablalblablbllbalbllball"""
        
        self.__warning_Label = Label(self.__root, text = warning_msg, font=('Helvetica', 16), pady = 1, padx = 1, bg = 'red')
        self.__warning_Label.grid(row = 1, column=0, columnspan = 3, sticky='nsew')
        
    def set_grid_settings(self):
        #self.__root.rowconfigure(0, weight=1)
        self.__root.rowconfigure(1, weight=0)
        
        self.__root.grid_columnconfigure(0, weight=2)
        self.__root.grid_columnconfigure(1, weight=2)
        self.__root.grid_columnconfigure(2, weight=1)
        
    def determine_use_case(self, key):
        if 'slot' in key:
            return 'slot'
        
        if key == 'COM_port':
            return 'COM_port'
        
        return 'parameter'

    def aplay_theme_pack(self, theme_pack: dict):
        self.__theme_pack = theme_pack
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        header = theme_pack['header']
        
        self.__root.config(bg = bg)
        self.__header_config.config(bg = header, fg = fg)
        
        for row in self.__row_obj_list:
            row: ConfigFileRow
            row.aplay_theme_pack(theme_pack=self.__theme_pack)
        
    def create_row_obejcts(self):
        self.__row_obj_list = []
        for index, (key, value) in enumerate(self.__data.items()):
            use_case = self.determine_use_case(key)
            row_obj = ConfigFileRow( root = self.__root, key = key, value = value,
                                            type = type(value), row_idx = index + 2, data = self.__data, 
                                            file_path = self.__config_directory_path, use_case= use_case, db = self.__db)
            self.__row_obj_list.append(row_obj)
        
class ConfigFileRow():
    __root: Frame
    __db: DB
    def __init__(self, root: Frame, file_path, key, value, type, row_idx, data, db, use_case):
        self.__key = key
        self.__value = value
        self.__type = type
        self.__root = root
        self.__row_idx = row_idx
        self.__data = data
        self.__use_case = use_case
        self.__file_path = file_path
        self.__db = db
        self.__db.get_slots()
        self.create_row()

    def create_row(self):
        self.__root.rowconfigure(self.__row_idx, weight = 1)
        self.create_label()
        self.create_entry()
        self.create_enter_button()

    def create_label(self):
        font = tkinter.font.Font(family = 'Times New Roman', size = 14)
        self.__row_label = Label(self.__root, text = str(self.__key), font=font, relief='solid',  pady = 5, padx = 1)
        self.__row_label.grid(row=self.__row_idx, column = 0, sticky='nsew', ipadx = len(str(self.__value)))

    def create_entry(self):
        entry_text = StringVar(value = self.__value)
        self.__row_entry = Entry(self.__root, textvariable = entry_text)
        self.__row_entry.grid(row = self.__row_idx, column = 1, sticky='nsew', ipadx = len(str(self.__value)))

    def create_enter_button(self):
        self.__enter_button = Button(self.__root, text = "Enter", command = self.enter_button_callback, font=('Helvetica', 12))
        self.__enter_button.grid(row = self.__row_idx, column = 2, sticky='nsew', ipadx = len('entry'))
        
    def aplay_theme_pack(self, theme_pack: dict):
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        
        self.__row_entry.config(bg = entry, fg = fg)
        self.__enter_button.config(bg = button, fg = fg)
        self.__row_label.config(bg = bg, fg = fg)
        
    def enter_button_callback(self):
        self.entry = self.__row_entry.get()
        print(f"entry: {self.entry}")
        print(f"value: {self.__value}")
        
        if not self.is_entry_vaild(entry = self.entry):
            self.__row_entry.delete(0, END)
            if self.__value is not None:
                self.__row_entry.insert(0, self.__value)
            return
        
        if self.entry is None:
            self.__row_entry.delete(0, END)
        
        if self.__use_case == 'slot':
            self.slot_entry_changes(old_slot = self.__value, entry = self.entry)
            
        self.__value = self.entry
        self.__data[self.__key] = self.__value
        self.write_into_json_file(self.__data)
        
        print(f"""
              used_slots: {self.used_slots()})
              """)
    
        
    def slot_entry_changes(self, old_slot, entry):
        print("do slot changes")
        if entry is None:
            self.db_delete_slot(old_slot)
            return
        
        if old_slot is None:
            self.db_insert_new_slot(slot = entry, role=1)
            
        else:
            
            self.db_update_slot(new_slot = self.entry, old_slot= self.__value)
   
    """
    """
    def is_entry_vaild(self, entry: str):
        if entry == self.__value:
            return False
        
        if self.__use_case == 'parameter':
            return self.is_entry_valid_parameter(entry)
            
        if self.__use_case == 'slot':
            return self.is_entry_valid_slot(entry)
        
        if self.__use_case == 'COM_port':
            return self.is_entry_valid_comport(entry)
        
    def is_entry_valid_parameter(self, entry: str):
        if entry.isdigit() and (not self.__type is int) and (not self.__type is NoneType):
            print("g")
            print(self.__type)
            return
        
        if not entry.isdigit() and self.__type is int:
            print("f")
            return
        
        if entry.isdigit() and self.__type is int:
            self.entry = int(entry)
            
        if entry == "":
            self.entry = 0
            
        return True
    
    def is_entry_valid_slot(self, entry):
        if entry == "":
            self.entry = None
            return True
            
        if not entry.isdigit():
            return
            
        if int(entry) < 1:
            return
        
        if int(entry) in self.used_slots():
            print("entry invalid")
            return

        self.entry = int(entry)
        print("entry valid")
        return True
        
    def is_entry_valid_comport(self, entry):
        #regular expression if the entry is for the COMPort: COM'int'
        comp = r"^COM\d+$" 
            
        if not bool(match(comp, entry)):
            print("invalid COMport")
            return
        
        return True
            
    """
    """
    def write_into_json_file(self, data):
        with open(self.__file_path / 'config.json', 'w') as file:
            json.dump(data, file, indent=4)
            
    """
    database actions
    """
    def db_delete_slot(self, slot):
        print('delete')
        self.__db.delete_row(slot = slot)
        
    def db_insert_new_slot(self, slot, role):
        print('insert')
        self.__db.insert_new_row(slot = slot, role = role)
        
    def db_update_slot(self, new_slot, old_slot):
        print('update')
        self.__db.update_slot(new_slot = new_slot, old_slot=old_slot)
        
    def used_slots(self):
        return self.__db.get_slots()


import time
from datetime import datetime
from tkinter import *
from tkinter.ttk import Progressbar
from cytomat_plus.db import DB
from .cytomat_plus import CytomatPlus
from cytomat.parameters import Parameters
from cytomat_plus.manager_settings import ManagerSettings


class GUI():
    __root: Tk
    __cytomat: CytomatPlus
    __DB: DB
    __manager_settings: ManagerSettings
    def __init__(self):
        self.__root = Tk()
        self.__root.title('Cytomat Gui')
        self.__root.geometry('1920x1080')

        self.__parameters = Parameters()
        self.__parameters.load()
        #self.__root.attributes('-fullscreen', True)

        self.__cytomat = CytomatPlus(self.__parameters.COM_port)
        self.__DB = DB()

        self.__manager_settings = ManagerSettings(db = self.__DB, root = self.__root)
        self.__manager_settings.create_overview()

        self.__interrupted: bool = False
        self.checker: bool = False
        self.busy_v: bool = False
        self.__pause: bool = False
        self.__root.protocol('WM_DELETE_WINDOW', self.on_close)
        self.busy_row: int = None
        #contains all slots where role = 0 in form of a dictionary 
        self.__db_data: list[dict] = self.__DB.get_rows_role_dict(role = 0)
        self.check_if_transfer_occupied()

    def get_root(self):
        return self.__root
    
    #sets all of rows in state busy
    def set_busy_state(self, busy: bool):
        for row in self.__list_of_row_objects:
            row.busy(busy)

    def get_cytomat_status(self):
        return self.__cytomat.get_status()

    """
    Creates Overview consisting of 4 headers and n rows x 6 columns of witch 3 are Data and 3 commands
    """
    def create_overview(self):
        self.set_dynamic_window_settings()
        self.create_headers()
        self.create_setting_button()
        self.create_row_objects()

    #every row in th gui consists of an object of the gui_row class
    def create_row_objects(self):
        self.__list_of_row_objects = []

        for row_idx, data_dict in enumerate(self.__db_data):
            row = GUI_row(row_idx+1, root=self.__root, gui = self, DB = self.__DB, cytomat=self.__cytomat, transfer_oc=self.checker, data = data_dict)
            self.__list_of_row_objects.append(row)

    #every row object is re initialzied
    def update_overview(self):
        self.__db_data = self.__DB.get_rows_role_dict()
        self.create_row_objects()

    #settings to scale the grid proportional to the window size  
    def set_dynamic_window_settings(self):
        self.__root.columnconfigure(0, weight=1)
        self.__root.columnconfigure(1, weight=1)
        self.__root.columnconfigure(2, weight=1)
        self.__root.columnconfigure(3, weight=1)
        self.__root.columnconfigure(4, weight=1)
        self.__root.columnconfigure(5, weight=1)

        for i in range(1, len(self.__db_data) + 1):
            self.__root.rowconfigure(i, weight = 1)
    
    def create_headers(self):
        y = 30
        self.__header_1 = Label(self.__root, text='SlotNumer', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_1.grid(column=0, row=0, pady = 20, sticky = 'ew', columnspan= 1, ipadx = len('SlotNumber'), ipady=y)

        self.__header_2 = Label(self.__root, text='Role', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_2.grid(column=1, row=0, pady = 20, sticky = 'ew', columnspan= 1, ipadx = len('Role'), ipady=y)

        self.__header_3 = Label(self.__root, text='AssignedPlatePlateId', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5 , font=('Helvetica', 16))
        self.__header_3.grid(column=2, row=0, pady = 20, sticky = 'new', columnspan= 1, ipadx= len('AssignedPlatePlateId'), ipady=y)

        self.__header_4 = Label(self.__root, text='command', bg= 'lightgray', relief='solid', borderwidth= 2, pady = 5, font=('Helvetica', 16) )
        self.__header_4.grid(column=3, row=0, pady = 20, sticky = 'new', columnspan= 3, ipadx = len('commands'), ipady=y)

    def create_setting_button(self):
        self.__setting_button = Button(self.__root, text = "Settings", command = self.setting_button_callback)
        self.__setting_button.grid(row = 0, column = 5, ipadx = 15, ipady = 15, sticky='ne', pady=5)

    def setting_button_callback(self):
        self.__manager_settings.display_window()

    """
    """

    """
    every 100ms the function calls itself as long the main window exists(without blocking the main thread) and checks if the transfer station state has changed. 
    If so the function updates the overview 
    """
    def check_if_transfer_occupied(self):
        if not self.__root.winfo_exists():
            return 
    
        transfer_station_occupied = self.get_cytomat_status().transfer_station_occupied
        if self.checker != transfer_station_occupied:
            self.checker = transfer_station_occupied
            self.update_overview()

        self.after_id = self.__root.after(25, self.check_if_transfer_occupied)
    
    """
    It kills all the backround operations to close the main window properly without raising an error(just check_if_transfer_occupied)
    """
    def on_close(self):
        if self.after_id is not None:
            self.__root.after_cancel(self.after_id)
        self.__root.destroy()

    """
    creates a window to insert new plates. It consits of a year, month, day and index Entry, wich can be in- or decresed by an plus or minus button.
    """
    def create_insert_new_plate_window(self, slot):
        self.__insert_window = Toplevel(master = self.__root)
        self.__insert_window.title('insert window')
        self.__insert_window.geometry('1000x400+400+400')

        self.__insert_window.columnconfigure(0, weight=1)
        self.__insert_window.columnconfigure(1, weight=1)
        self.__insert_window.columnconfigure(2, weight=1)
        self.__insert_window.columnconfigure(3, weight=1)
        self.__insert_window.columnconfigure(4, weight=1)
        self.__insert_window.columnconfigure(5, weight=1)

        self.__insert_window.rowconfigure(0, weight=1)
        self.__insert_window.rowconfigure(1, weight=1)
        self.__insert_window.rowconfigure(2, weight=1)
        self.__insert_window.rowconfigure(2, weight=1)

        self.year = StringVar()
        self.month = StringVar()
        self.day = StringVar()
        self.index_ = StringVar()

        current_year = datetime.now().year
        current_month = datetime.now().month
        current_day = datetime.now().day

        self.year.set(f'{current_year:02}')
        self.month.set(f'{current_month:02}')
        self.day.set(f'{current_day:02}')
        self.index_.set('a')

        year_label = Label(master = self.__insert_window, text= 'year', font = ('Helvetica', 14), relief='solid')
        month_label = Label(master = self.__insert_window, text = 'month', font = ('Helvetica', 14),relief='solid')
        day_label = Label(master = self.__insert_window, text = 'day', font = ('Helvetica', 14), relief='solid')
        index_label = Label(master = self.__insert_window, text ='index', font = ('Helvetica', 14), relief='solid')

        year_label.grid(row = 1, column = 0, ipady = 20, sticky = 'ew')
        month_label.grid(row = 1, column = 1, ipady = 20, sticky = 'ew')
        day_label.grid(row = 1, column = 2, ipady = 20, sticky = 'ew')
        index_label.grid(row = 1, column = 3, ipady = 20, sticky = 'ew')

        entry_year = Entry(master = self.__insert_window, textvariable = self.year,  state = 'readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_month = Entry(master =self.__insert_window, textvariable =self.month,  state='readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_day = Entry(master =self.__insert_window, textvariable = self.day, state='readonly', relief = 'solid', font = ('Helvetica', 14))
        entry_idx = Entry(master =self.__insert_window,textvariable = self.index_, state = 'readonly',relief = 'solid', font = ('Helvetica', 14))

        entry_year.grid(row = 2, column = 0, ipady = 40, ipadx = 80)
        entry_month.grid(row = 2, column = 1,ipady = 40, ipadx = 80)
        entry_day.grid(row = 2, column = 2, ipady = 40, ipadx = 80)
        entry_idx.grid(row = 2, column = 3,ipady = 40, ipadx= 80)

        submit_btn = Button(master =self.__insert_window, text = "Submit",command = lambda: self.submit_btn_callback(slot), font = ('Helvetica', 12), relief = 'solid')
        submit_btn.grid(row=0, column=4, ipadx = 80, ipady = 40, rowspan=4, sticky='ns')

        year_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.year_change(0), font = ('Helvetica', 20), relief = 'solid')
        year_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.year_change(1), font = ('Helvetica', 20), relief = 'solid')
        year_btn_plus.grid(row = 0, column=0, sticky = 'ew', ipady = 40)
        year_btn_minus.grid(row = 3, column=0, sticky = 'ew', ipady = 40)

        month_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.month_change(0), font = ('Helvetica', 20), relief = 'solid')
        month_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.month_change(1), font = ('Helvetica', 20), relief = 'solid')
        month_btn_plus.grid(row = 0, column = 1, sticky = 'ew', ipady = 40)
        month_btn_minus.grid(row = 3, column = 1, sticky = 'ew', ipady = 40)

        day_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.day_change(0), font = ('Helvetica', 20), relief = 'solid')
        day_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.day_change(1), font = ('Helvetica', 20), relief = 'solid')
        day_btn_plus.grid(row = 0, column = 2, sticky = 'ew', ipady = 40)
        day_btn_minus.grid(row = 3, column = 2, sticky = 'ew', ipady = 40)

        idx_btn_plus = Button(master =self.__insert_window, text = '+', command = lambda: self.idx_change(0), font = ('Helvetica', 20), relief = 'solid')
        idx_btn_minus = Button(master =self.__insert_window, text = '-', command = lambda: self.idx_change(1), font = ('Helvetica', 20), relief = 'solid')
        idx_btn_plus.grid(row = 0, column = 3, sticky = 'ew', ipady = 40)
        idx_btn_minus.grid(row=3, column = 3, sticky = 'ew', ipady = 40)

        self.wait_for_submit = IntVar()
        self.submit_insert = False
        self.__insert_window.protocol('WM_DELETE_WINDOW',self.on_close_insert_window)

        self.__insert_window.update()
        self.__insert_window.update_idletasks()

    def on_close_insert_window(self):
        self.submit_insert = False
        self.close_insert_window()
        self.wait_for_submit.set(1)

    def close_insert_window(self):
        self.__insert_window.destroy()

    def submit_btn_callback(self, slot: int):
        plate_id = self.year.get() + self.month.get() + self.day.get() + self.index_.get()

        if self.__DB.is_plate_assigned(plate_id):
            print("This Plate is allready assigned to a Slot")
            return 
        self.__DB.assigne_plate(plate_id=plate_id, slot = slot)
        self.close_insert_window()
        self.update_overview()
        self.submit_insert = True
        self.wait_for_submit.set(1)

    """
    p_o_m stands for plus or minus, 0 is plus, 1 is minus
    """
    def year_change(self, p_o_m: int):
        curr_year = int(self.year.get())
        if p_o_m == 0:
            new_year = curr_year + 1
        else:
            new_year = curr_year - 1

        self.year.set(f'{new_year:02}')

    def month_change(self, p_o_m: int):
        curr_month = int(self.month.get())
        if p_o_m == 0:
            new_month = curr_month +1
        else:
            new_month = curr_month -1

        if new_month > 12:
            new_month = 1

        if new_month < 1:
            new_month = 12
        self.month.set(f'{new_month:02}')

    def day_change(self, p_o_m: int):
        curr_day = int(self.day.get())
        if p_o_m == 0:
            new_day = curr_day +1 
        else:
            new_day = curr_day -1

        if new_day >31:
            new_day = 1

        if new_day <1:
            new_day =31

        self.day.set(f'{new_day:02}')
        
    def idx_change(self, p_o_m: int):
        curr_index = self.index_.get()
        curr_ascii = ord(curr_index)
        if p_o_m == 0:
            new_ascii = curr_ascii + 1
        else:
            new_ascii = curr_ascii - 1
        
        if new_ascii > ord('z'):
            new_index = 'a'
        elif new_ascii < ord('a'):
            new_index = 'z'
        else:
            new_index = chr(new_ascii)

        self.index_.set(new_index)

    """
    Creates a progress window with opens when a command has been clicked. Consisting of a progress bar, status label and an interrupt button
    """
    def create_progress_window(self):
        self.__progress_window = Toplevel(master = self.__root)
        self.__progress_window.geometry('800x200+400+400')
        self.__progress_window.title('Progress window')
        #self.__progress_window.overrideredirect(True)

        self.__progressbar = Progressbar(self.__progress_window, maximum=100)
        self.__progressbar.grid(row = 0, column=0, columnspan=2, padx=5, pady = 50, sticky = 'new', ipadx = 200)
        
        self.__status_label = Label(self.__progress_window, text = "status")
        self.__status_label.grid(row = 0, column=1, columnspan=2, ipadx = 200, pady = 50, sticky = 'sew')
        
        self.__pause_btn = Button(self.__progress_window, text="Pause", bg = 'red', command = self.pause_btn_callback)
        self.__pause_btn.grid(row = 0, column = 3, ipadx = 100, ipady = 50, pady = 20, padx=5)

        self.__progress_window.update()

    def close_progress_window(self):
        self.__progress_window.destroy()

    def pause_btn_callback(self):
        self.__pause = True
        self.create_interrupt_window()
        print("set .__pause to True")

    def get_interrupted(self):
        return self.__interrupted

    def set_interrupted(self, vbool):
        self.__interrupted = vbool

    def update_progressbar(self, percent: float):
        self.__progressbar['value'] = percent
        self.__root.update_idletasks()
    
    def update_status_label(self, text: str):
        self.__status_label.config(text = text)
        self.__root.update()

    def is_progress_window_exists(self)->bool:
        return self.__progress_window.winfo_exists()

    """
    A loop with checks if the Queue, wich the main thread and the "Cytomat execute thread" are using to communicate,
    has been updated and then Updates the Progress window
    """
    def start_queue_checker(self):
        self.__status_queue.set_initial_queue_conditions()
        self.check_status_queue()

    def wait_for_queue_checker_to_end(self):
        self.__queue_checker_thread.join()

    #Gets a list of commands and gives them each to the Cytomat.execute, to execute them.
    def task_manager(self, task_list)->bool:
        print("task manager")
        task_completed = None
        for idx, command in enumerate(task_list):
            self.current_idx = idx
            if self.__pause is True:
                self.__pause = False
                self.__root.wait_variable(self.__interrupt_window_submit)

                if self.interrupt_state != 'continue':
                    task_completed = False
                    return print(task_completed)

            error = self.__cytomat.execute(command, timeout=120)

            if error != None:
                print(error)
                print("error")
                self.__cytomat.execute_error = None
                self.error_handler(error)
                return

            if self.is_progress_window_exists():
                self.update_progressbar(idx/len(task_list) * 100)
                self.update_status_label(command.__name__)
                self.__progress_window.update()
            time.sleep(1)

        print("done")
        task_completed = True
        return print(task_completed)

    def error_handler(self, error):
        print(error)
        self.error_raised = True


    """
    Creates a window to Initialize the Plate handler of the Cytomat after an command has been interrupted 
    """
    def create_interrupt_window(self):
        self.__reboot_window = Toplevel(master = self.__root)
        self.__reboot_window.title('Interrupt window')
        self.__reboot_window.geometry('600x350+400+400')
        self.__reboot_window.overrideredirect(True)

        self.__interrupt_window_submit = IntVar()

        self.__cyto_init_btn = Button(self.__reboot_window, bg = 'red', text="Re-Init Cytomat PH", command = self.callback_cyto_init_btn, relief='solid', font=('Helvetica', 16))
        self.__cyto_init_btn.pack(expand=True, fill='both')

        self.__continue_btn = Button(self.__reboot_window, text="Continue", command = self.callback_continue_btn, font =('Helvetica', 16))
        self.__continue_btn.pack(expand=True, fill='both')

        self.__get_back_btn = Button(self.__reboot_window, text = "Undo command", command=lambda: self.callback_get_back_btn(), font =('Helvetica', 16))
        self.__get_back_btn.pack(expand = True, fill = 'both')

        self.__reboot_window.update()
        self.__reboot_window.update_idletasks()

    def callback_get_back_btn(self):
        self.interrupt_state = ['get_back', self.current_idx]
        self.__interrupted = True
        self.__reboot_window.destroy()
        self.__interrupt_window_submit.set(1)

    def callback_cyto_init_btn(self):
        self.interrupt_state = ['reinitialize']

        self.__interrupted = True
        self.__reboot_window.destroy()
        self.__interrupt_window_submit.set(1)

    def callback_continue_btn(self):
        self.interrupt_state = 'continue'
        self.__interrupt_window_submit.set(1)
        self.__reboot_window.destroy()
    """
    """
    def create_interrupt_actions_window(self):
        self.__display_interrupt_acction = Toplevel(master =self.__root)
        self.__display_interrupt_acction.geometry('600x350+400+400')
        self.__display_interrupt_acction.title('Interrupt Action')

        self.__display_action_label = Label(master=self.__display_interrupt_acction, text = "currently doing:")
        self.__display_action_label.pack(expand=True, fill='both')

    def modifiy_display_label(self,text_):
        self.__display_action_label.config(text=text_)
        self.__display_interrupt_acction.update()

    def destroy_display_interrupt(self):
        self.__display_interrupt_acction.destroy()
    
    def create_submit_delete_window(self, plate: str):
        self.submit_delete_wait = IntVar()

        self.__submit_delete_window = Toplevel(master=self.__root)
        self.__submit_delete_window.geometry('600x350+400+400')
        self.__submit_delete_window.title('Submit Delete')
        self.__submit_delete_window.overrideredirect(True)

        self.__submit_delete_window.columnconfigure(0, weight=1)
        self.__submit_delete_window.rowconfigure(0, weight=1)
        self.__submit_delete_window.rowconfigure(1, weight=1)
        self.__submit_delete_window.rowconfigure(2, weight=1)
        
        self.__submit_delete_label = Label(master= self.__submit_delete_window, text=f"To delete plate:{plate} press button", font = ('Helvetica', 14))
        self.__submit_delete_label.grid(row =0, column=0, sticky='nswe')

        self.__subimt_delete_btn = Button(master = self.__submit_delete_window, command=self.submit_delete_callback, text = "Commit Delete")
        self.__subimt_delete_btn.grid(row=1, column=0)
        
        self.__cancel_delete = Button(master = self.__submit_delete_window, command=self.cancel_delete, text = "Cancel")
        self.__cancel_delete.grid(row=3, column=0)

    def submit_delete_callback(self):
        self.submit_delete = True
        self.submit_delete_wait.set(1)
        self.__submit_delete_window.destroy()

    def cancel_delete(self):
        self.submit_delete = False
        self.submit_delete_wait.set(1)
"""
Every row of the gui consists of an object of the GUI_row class, a 6 collumns of wich 3 are data and 3 commands
"""
class GUI_row():
    __label_slot    : Label
    __label_role    : Label
    __label_plate   : Label
    __btn_media     : Button
    __btn_eject     : Button
    __btn_messurement   : Button
    __DB            : DB
    cy              : CytomatPlus
    row_idx         : int
    command_map: dict
    #db_row: DB_row

    def __init__(self, row_idx, root, gui, DB, cytomat, transfer_oc: bool, data: dict):
        self.__root: Tk = root
        self.cy: CytomatPlus = cytomat
        self.gui: GUI = gui
        self.__DB = DB

        self.row_idx = row_idx

        self.slot = data['slot']
        self.role = data['role']
        self.plate = data['plate']
        self.occupied = data['occupied']

        self.transfer_oc = transfer_oc
        self.cyto_status = self.gui.get_cytomat_status()
        self.command_map = self.command_map_()
        self.create()

    def create(self):
        if self.plate is None:
            plate_bool = False
        else: 
            plate_bool = True

        if self.occupied == 0:
            occupied_bool = False
        else:
            occupied_bool = True

        command_map = self.command_map[(self.transfer_oc, plate_bool, occupied_bool)]
        
        cmd_left =      command_map['cmd_left']
        bg_left =       command_map['color_left']
        text_left =     command_map['text_left']
        state_left =    command_map['state_left']

        cmd_middle =    command_map['cmd_middle']
        bg_middle =     command_map['color_middle']
        text_middle =   command_map['text_middle']
        state_middle =  command_map['state_middle']

        cmd_right =     command_map['cmd_right']
        bg_right =      command_map['color_right']
        text_right =    command_map['text_right']
        state_right =   command_map['state_right']
        
        self.__btn_media = Button(self.__root, text = text_left, bg = bg_left, command= cmd_left, state = state_left, font=('Helvetica', 14))
        self.__btn_eject = Button(self.__root, text = text_middle, bg = bg_middle, command = cmd_middle, state = state_middle, font=('Helvetica', 14))
        self.__btn_messurement = Button(self.__root, text = text_right, bg = bg_right, command = cmd_right, state = state_right, font=('Helvetica', 14))

        self.__label_slot =Label(self.__root, text=self.slot , relief='solid',  pady = 1, padx = 1, font=('Helvetica', 14))
        self.__label_role =Label(self.__root, text=self.role , relief='solid',  pady = 1, padx = 1, font=('Helvetica', 14))
        self.__label_plate=Label(self.__root, text=self.plate, relief='solid', pady = 1, padx = 1, font=('Helvetica', 14))

        if self.gui.busy_v is True:
            self.busy(True)
            
        if self.gui.busy_row == self.row_idx:
            self.busy_row(True)

        if self.role == 1:
            self.busy(True)

        self.__label_slot.grid(  column = 0, row = self.row_idx, sticky='nsew', pady = 1)
        self.__label_role.grid(  column = 1, row = self.row_idx, sticky='nsew', pady = 1)
        self.__label_plate.grid( column = 2, row = self.row_idx, sticky='nsew', pady = 1)
        self.__btn_media.grid(   column=  3, row = self.row_idx, sticky = 'nsew', ipadx = len('Mediachange'))
        self.__btn_eject.grid(   column=  4, row = self.row_idx, sticky = 'nsew', ipadx = len('Insert'))
        self.__btn_messurement.grid(    column=  5, row = self.row_idx, sticky = 'nsew', ipadx = len('Test'))

    # The three bools in the hashmap stand for 1.: Transfer_station Occupied, 2.: Plate is assigned, 3.: the Slot is occupied 
    def command_map_(self)->dict:
        cmd_map = {
            (False, False, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_register(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Register',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (False, False, True): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'grey', 'text_left' :'None',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'None',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'grey', 'text_right': 'None'},

            (False, True, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_delete(), 'state_middle': 'normal', 'color_middle': 'yellow', 'text_middle':'Delete',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (False, True, True): {'cmd_left': lambda: self.callback_mediachange(), 'state_left': 'normal', 'color_left': 'green', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_eject(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Eject',
                                    'cmd_right': lambda: self.callback_messurement(), 'state_right': 'normal', 'color_right': 'green', 'text_right': 'FOC48'},

            (True, False, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_register(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Register',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (True, False, True): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'grey', 'text_left' :'None',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'None',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'grey', 'text_right': 'None'},

            (True, True, False): {'cmd_left': None, 'state_left': 'disabled', 'color_left': 'white', 'text_left' :'mediachange',
                                    'cmd_middle': lambda: self.callback_insert(), 'state_middle': 'normal', 'color_middle': 'green', 'text_middle':'Insert',
                                    'cmd_right': None, 'state_right': 'disabled', 'color_right': 'white', 'text_right': 'FOC48'},

            (True, True, True): {'cmd_left': lambda: self.callback_mediachange(), 'state_left': 'normal', 'color_left': 'green', 'text_left' :'mediachange',
                                    'cmd_middle': None, 'state_middle': 'disabled', 'color_middle': 'grey', 'text_middle':'Insert',
                                    'cmd_right': lambda: self.callback_messurement(), 'state_right': 'normal', 'color_right': 'green', 'text_right': 'FOC48'},
            }
        return cmd_map
    
    def busy(self, lock: bool): 
        if (lock) is True:
            self.__label_slot.config( bg = 'grey')        
            self.__label_role.config( bg = 'grey')        
            self.__label_plate.config(bg = 'grey')        
            self.__btn_media.config(state = DISABLED, bg = 'grey')
            self.__btn_eject.config(state = DISABLED, bg = 'grey')
            self.__btn_messurement.config( state = DISABLED, bg = 'grey')
        else:
            self.__label_slot.config(bg = 'white')      
            self.__label_role.config(bg = 'white')        
            self.__label_plate.config(bg = 'white')        
            self.__btn_media.config(state = NORMAL)
            self.__btn_eject.config(state = NORMAL)
            self.__btn_messurement.config( state = NORMAL)
        self.__root.update()

    def busy_row(self, lock: bool):
        if (lock) is True:
            self.__label_slot.config(bg= 'green')
            self.__label_role.config(bg= 'green')
            self.__label_plate.config(bg= 'green') 
        else:
            self.__label_slot.config(bg= 'white')
            self.__label_role.config(bg= 'white')
            self.__label_plate.config(bg= 'white')
        self.__root.update()

    """
    Callback functions of the Buttons
    """

    """
    checks if the busy is not busy, the assigned role is the right one and if the plate is not assigned allready.
    Then sets the Gui to busy and 
    """
    def callback_register(self):
        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storrage")
            return

        if self.plate is not None:
            print("The Slot is allready assigned to a Plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_insert_new_plate_window(self.slot)
        self.__root.wait_variable(self.gui.wait_for_submit)
        
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

        """
        Delete logics
        """
    def callback_delete(self):
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)

        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_submit_delete_window(plate = self.plate)
        self.__root.wait_variable(self.gui.submit_delete_wait)

        if self.gui.submit_delete is True:
            self.delete_plate()

        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()


    def delete_plate(self):
        self.__DB.delete_plate(self.slot)
        print('deleted')

    """
    insert logics
    """
    def callback_insert(self):
        status = self.gui.get_cytomat_status()

        #Basic checks, looking if the Plate is in the right state:
        if status.transfer_station_occupied is False:
            print("There is no plate on the transfer station")
            return

        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storrage")
            return

        #Blocking all the other buttons, Gui is busy now
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)

        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager([lambda: self.cy.plate_handler.move_plate_from_transfer_station_to_slot(self.slot)])

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            self.gui.set_interrupted(False)
            self.gui.create_interrupt_actions_window()

            self.choose_interrupt_action(key = 'insert')
        else:
            self.__DB.update_occupied(self.slot, 1)

        self.__root.update()
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def insert(self):
        self.cy.execute(lambda: self.cy.plate_handler.move_plate_from_transfer_station_to_slot(self.slot), timeout = 20)
 
    def invert_insert_list(self, current_idx: int)->list:
        inverted_list = self.cy.plate_handler.inverted_move_plate_from_transfer_station_to_slot_v2(self.slot)
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list
    """
    eject logics
    """
    def callback_eject(self):
        status = self.gui.get_cytomat_status()

        if status.transfer_station_occupied is True:
            print("The transfer station is occupiued")
            return

        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("This Plate is not for storrage")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager(task_list = [lambda: self.cy.plate_handler.move_plate_from_slot_to_transfer_station(self.slot)])

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            self.gui.create_interrupt_actions_window()

            self.gui.set_interrupted(False)

            self.choose_interrupt_action(key = 'eject')
            self.gui.destroy_display_interrupt()

        else:
            self.__DB.update_occupied(self.slot, 0)    #TODO update in re_init_window 

        self.__root.update()
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def eject(self):
        print("running eject")
        self.cy.execute(lambda: self.cy.plate_handler.move_plate_from_slot_to_transfer_station(self.slot), timeout = 20)
        print("eject Done")

    def invert_eject_list(self, current_idx: int)->list:
        inverted_list = self.cy.plate_handler.inverted_move_plate_from_slot_to_transfer_station_v2(self.slot)
        modified_invert_list = []
        for i in range(current_idx+1):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    """
    mediachange logics
    """
    def callback_mediachange(self):
        if self.gui.busy_v is True:
            return

        if self.role != 0:
            print("his Plate is not for mediachange")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()

        self.gui.task_manager(task_list = self.cy.plate_handler.test3())
        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            """if self.gui.error_raised:
                print("Error raised")"""

            self.gui.set_interrupted(False)
            self.gui.create_interrupt_actions_window()

            self.choose_interrupt_action(key='mediachange')
            self.gui.destroy_display_interrupt()

        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def invert_mediachange_list(self, current_idx):
        inverted_list = self.cy.plate_handler.inverted_test3()
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    """
    messurement logics
    """
    def callback_messurement(self):
        #self.__DB.delete_plate(row_idx=self.row_idx-1)
        if self.gui.busy_v is True:
            print("Cyto is busy")
            return

        if self.role != 0:
            print("This Plate is not for mediachange")
            return

        if self.plate is None:
            print("This slot is not assigned to a plate")
            return

        if self.cyto_status.shovel_occupied is True:
            print("shovel occupied")
            return
        
        self.gui.busy_v = True
        self.gui.busy_row = int(self.row_idx)
        self.gui.set_busy_state(True)
        self.busy_row(True)

        self.gui.create_progress_window()
        self.gui.task_manager(self.messurement())

        self.gui.close_progress_window()

        if self.gui.get_interrupted() is True:
            
            self.gui.set_interrupted(False)

            self.gui.create_interrupt_actions_window()
            self.choose_interrupt_action(key='measurement')
            self.gui.destroy_display_interrupt()
                
            """self.__DB.write_log(time_stamp = datetime.now(), action= '', plate_id=self.plate)"""
  
        self.gui.busy_v = False
        self.gui.busy_row = None
        self.gui.update_overview()

    def messurement(self):
        return self.cy.plate_handler.get_plate_do_measurement_bring_back_plate(plate = self.plate, slot_a = self.slot)

    def invert_measurement_list(self, current_idx):
        inverted_list = self.cy.plate_handler.inverted_get_plate_do_measurement_bring_back_plate(slot_a = self.slot)
        modified_invert_list = []
        for i in range(current_idx):
            modified_invert_list.append(inverted_list[i])

        return modified_invert_list

    def init_plate_handler(self):
        self.cy.execute(self.cy.plate_handler.initialize)

    def choose_interrupt_action(self, key):
        command_mapping = {'mediachange': self.invert_mediachange_list,
                           'eject': self.invert_eject_list,
                           'insert': self.invert_insert_list,
                           'measurement': self.invert_measurement_list}

        match self.gui.interrupt_state[0]:
                case 'initialize':
                    self.gui.modifiy_display_label(text_ = "Currently doning: Initializing plate handler")
                    self.gui.task_manager([self.init_plate_handler])
                case 'get_back':
                    self.gui.modifiy_display_label(text_ = "Currently doning: Reversing the acction")
                    modified_invert_list = command_mapping[key](self.gui.interrupt_state[1])
                    self.gui.task_manager(task_list=modified_invert_list)
                case _: 
                    pass


import sqlite3
from sqlite3 import IntegrityError

class DB():
    def __init__(self):
        self.conn = sqlite3.connect('C:\ProgramData\Cytomat\slots.db', timeout=10)
        self.conn.execute('PRAGMA journal_mode=WAL')
        self.data_db = self.get_SlotNumber_data()
    """
    getter 
    """
    def get_column_names(self):
        column_names = ['SlotNumber', 'Role', 'AssignedPlatePlateId']
        return column_names

    def get_SlotNumber_data(self)-> list[list]:
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber, Role, AssignedPlatePlateId, Occupied FROM slots ORDER BY SlotNumber ASC')
        data = cursor.fetchall()
        cursor.close()
        return data
        
    def get_slots(self):
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber FROM slots ORDER BY SlotNumber ASC')
        data = [row[0] for row in cursor.fetchall()]
        cursor.close()
        return data

    def get_usable_rows(self, role: int = 0):
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber, Role, AssignedPlatePlateId, Occupied FROM slots WHERE role = ? AND AssignedPlatePlateId IS NOT NULL ORDER BY SlotNumber ASC', (role, ))
        data = cursor.fetchall()
        cursor.close()
        return data
        
    def get_roled_slots(self):
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber FROM slots WHERE NOT role = 0 ORDER BY SlotNumber ASC')
        data = [row[0] for row in cursor.fetchall()]
        cursor.close()
        return data
    
    def get_unroled_slots(self):
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber FROM slots WHERE role = 0 ORDER BY SlotNumber ASC')
        data = [row[0] for row in cursor.fetchall()]
        cursor.close()
        return data
    """
    gets all rows, depending on role and converts the list of tupels into a list of dictionarys and returns it
    """
    def get_rows_role_dict(self, role: int = 0):
        cursor = self.conn.cursor()
        cursor.execute('SELECT SlotNumber, Role, AssignedPlatePlateId, Occupied FROM slots WHERE Role = ? ORDER BY SlotNumber ASC', (role, ))
        data_tupels = cursor.fetchall()
        data_dicts = []
        for tupel in data_tupels:
            row_dict = {}
            row_dict['slot'] = tupel[0]
            row_dict['role'] = tupel[1]
            row_dict['plate'] = tupel[2]
            row_dict['occupied'] = tupel[3]
            data_dicts.append(row_dict)

        cursor.close()
        return data_dicts
        
    def slot(self, row_idx: int)->int:
        return self.data_db[row_idx][0]
    
    def role(self, row_idx: int)->int:
        return self.data_db[row_idx][1]
    
    def plate(self, row_idx: int)->str:
        return self.data_db[row_idx][2]

    def occupied(self, row_idx: int)->int:
        return self.data_db[row_idx][3]
    
    """
    setter
    """
    def insert_new_row(self, slot, role: int = 0, plate: str = None, occupied: int = 0, current_plate_id: str = None):
        cursor = self.conn.cursor()
        try:
            cursor.execute('INSERT INTO slots (SlotNumber, Role, AssignedPlatePlateId, Occupied, CurrentPlateId) VALUES (?, ?, ?, ?, ?)', (slot, role, plate, occupied, current_plate_id))
            self.conn.commit()
            
        except IntegrityError as e:
            print(e)

        finally:
            cursor.close()

    def assigne_slot(self, slot: int):
        cursor = self.conn.cursor()
        try:
            cursor.execute('INSERT INTO slots (SotNumber) VALUES (?)',(slot))
            self.conn.commit()
            
        except IntegrityError as e:
            print(e)

        finally:
            cursor.close()
            
    def update_slot(self, new_slot: int, old_slot: int):
        cursor = self.conn.cursor()
        try:
            cursor.execute('UPDATE slots SET SlotNumber = ? WHERE SlotNumber = ?', (new_slot, old_slot))
            self.conn.commit()
            
        except IntegrityError as e:
            print(e)

        finally:
            cursor.close()

    def assigne_plate(self, slot: int, plate_id:str):
        #slot = self.slot(row_idx=row_idx)
        cursor = self.conn.cursor()
        try:    
            cursor.execute('UPDATE slots SET AssignedPlatePlateId = ? WHERE SlotNumber = ?', (plate_id, slot))
            self.conn.commit()
        
        except IntegrityError as e:
            print(e)

        finally:
            cursor.close()
        
    def set_role(self, role: int, slot: int):
        cursor = self.conn.cursor()
        cursor.execute('UPDATE slots SET Role 0 ? WHERE SlotNumber = ?', (role, slot))
        self.conn.commit()
        cursor.close()
            
    def update_occupied(self, slot: int, occupied: int):
        cursor = self.conn.cursor()
        cursor.execute('UPDATE slots SET Occupied = ? WHERE SlotNumber = ?', (occupied, slot))
        self.conn.commit()
        cursor.close()
        
    """
    delete
    """
    def clear_entire_table(self):
        cursor = self.conn.cursor()
        cursor.execute('DELETE FROM slots')
        self.conn.commit()
        cursor.close()
        
    def delete_plate(self, slot):
        cursor = self.conn.cursor()
        cursor.execute('UPDATE slots SET AssignedPlatePlateId = NULL WHERE SlotNumber = ?', (slot, ))
        self.conn.commit()
        cursor.close()
    
    def delete_row(self, slot):
        cursor = self.conn.cursor()
        cursor.execute('DELETE FROM slots WHERE SlotNumber = ?', (slot, ))
        self.conn.commit()
        cursor.close()

    """
    rest
    """
    def is_plate_assigned(self, plate_id)->bool:
        cursor = self.conn.cursor()
        cursor.execute('SELECT 1 FROM slots WHERE AssignedPlatePlateId = ?',(plate_id,))
        result = cursor.fetchone()
        cursor.close()
        if result:
            return True
        else:
            return False

    def write_log(self, plate_id: str, action: str, time_stamp):
        cursor = self.conn.cursor()
        cursor.execute('INSERT INTO PlateActions (TimeStamp, PlateId, Acction) Values (?,?,?),',(time_stamp, plate_id, action,))
        self.conn.commit()
        cursor.close()


from tkinter import *
from cytomat_plus.db import DB

class ManagerSlots():
    __root: Frame
    __db: DB
    def __init__(self, root, db):
        self.__root = root
        self.__db = db
        self.__row_obj_list = []
        
    def create_overview(self):
        self.set_grid_settings()
        self.create_header()
        self.create_entry_area()
    
    def create_header(self):
        self.__header_slots = Label(self.__root, text = "Set the Slots wich should be used", font = ('Helvetica', 20), pady = 1, padx = 1, relief='solid')
        self.__header_slots.grid(row = 0, column=0, columnspan = 4, sticky='nsew', ipady = 30)
        
    def create_entry_area(self):
        y = 12
        
        self.__entry_bottom = Entry(self.__root, textvariable = 'bottom', justify = 'center')
        self.__entry_bottom.grid(row = 1, column = 0, sticky = 'nsew', ipady = y)
        
        self.__middle_label = Label(self.__root, text = '-', font=('Helvetica', 14))
        self.__middle_label.grid(row= 1, column = 1, sticky = 'nsew', ipady = y)
        
        self.__entry_upper = Entry(self.__root, textvariable = 'upper', justify = 'center')
        self.__entry_upper.grid(row = 1, column = 2, sticky = 'nsew', ipady = y)
    
        self.__enter_button = Button(self.__root, text = 'Enter', command = self.enter_button_callback)
        self.__enter_button.grid(row = 1, column = 3, sticky = 'nsew', ipady = y)
    
    def set_grid_settings(self):
        #self.__root.rowconfigure(0, weight=2)
        #self.__root.rowconfigure(1, weight=1)
        self.__root.rowconfigure(2, weight=1)
        self.__root.rowconfigure(3, weight=2)
        self.__root.rowconfigure(4, weight=2)
        self.__root.columnconfigure(0, weight = 2)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.columnconfigure(2, weight = 2)
        self.__root.columnconfigure(3, weight = 1)
        
    def aplay_theme_pack(self, theme_pack: dict):
        self.__theme_pack = theme_pack
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        header = theme_pack['header']
        
        self.__root.config(bg = bg)
        self.__header_slots.config(bg=header, fg = fg)
        self.__middle_label.config(bg = bg, fg = fg)
        self.__entry_upper.config(bg = entry, fg = fg)
        self.__entry_bottom.config(bg = entry, fg = fg)
        self.__enter_button.config(bg = button, fg = fg)
        
        for row in self.__row_obj_list:
            row: SlotsRow
            row.aplay_theme_pack(theme_pack=self.__theme_pack)
        
    def enter_button_callback(self):
        entry_bottom = self.__entry_bottom.get()
        entry_upper = self.__entry_upper.get()
        self.__entry_bottom.delete(0, END)
        self.__entry_upper.delete(0, END)
        
        if not self.is_entry_valid(entry_bottom, entry_upper):
            return
        
        entry_bottom = int(entry_bottom)
        entry_upper = int(entry_upper)
        
        kind_of_entry = self.kind_of_entry(entry_bottom, entry_upper)
         
    
    def is_entry_valid(self, entry_bottom: str, entry_upper: str):
        if not entry_bottom.isdigit() or not entry_upper.isdigit():
            return
        
        if int(entry_upper) < int(entry_bottom):
            return
        
        print("entry valid")
        return True
    
    def kind_of_entry(self, entry_bottom: int, entry_upper: int):
        pass
    
    def merge_entries():
        pass
    
class SlotsRow():
    def __init__(self) -> None:
        pass
    
    def aplay_theme_pack(self, theme_pack):
        pass


import subprocess
import sys
import time
import os
from cytomat.plate_handler import PlateHandler
from cytomat.status import OverviewStatus
from cytomat.serial_port import SerialPort
from cytomat.parameters import Parameters
from cytomat.convert_steps import ConvertSteps as CS

class PlateHandlerPlus(PlateHandler):

    def __init__(self, serial_port: SerialPort) -> None:
        self.parameters = Parameters()
        self.parameters.load()
        super().__init__(serial_port)
        self.lid_holder_slot = self.parameters.lid_holder_slot
        self.pipet_station_slot = self.parameters.pipet_station_slot
        self.max_steps_shovel = self.parameters.max_steps_shovel
        self.measurement_slot = self.parameters.measurement_slot
              
    def move_x_to_slot(self, slot: int) -> OverviewStatus:
        """time
        Move along to the given slot (only moves along the x axis)

        Parameters
        ----------
        slot
            The target slot
        """
        return self._PlateHandler__serial_port.issue_action_command(f"ll:xp {slot:03}")
    
    def leave_lid_at_holder(self) -> list[OverviewStatus]:
        if self.parameters.lid_holder_slot != None:
            command_list: list[OverviewStatus]
            command_list = [
                lambda: self.rotate_handler_to_slot(self.lid_holder_slot),
                lambda: self.move_x_to_slot(self.lid_holder_slot),
                lambda: self.move_handler_above_slot_height(self.lid_holder_slot),
                lambda: self.extend_shovel(),
                lambda: self.run_height_in_relative_mm(-15),
                lambda: time.sleep(2),
                lambda: self.retract_shovel(),
                lambda: self.reset_handler_position()]
            return command_list
        else:
            return print("Cup holder not initialized, missing arg type int at the initialisation of Object type CytomatPlus") 
    
    def get_lid_from_holder(self)-> list[OverviewStatus]:
        if self.parameters.lid_holder_slot != None:
            command_list: list[OverviewStatus]
            command_list = [
                lambda: self.rotate_handler_to_slot(self.lid_holder_slot),
                lambda: self.move_x_to_slot(self.lid_holder_slot),
                lambda: self.move_handler_below_slot_height(self.lid_holder_slot),
                lambda: self.extend_shovel(),
                lambda: self.run_height_in_relative_mm(15),
                lambda: time.sleep(2),
                lambda: self.retract_shovel(),
                lambda: self.reset_handler_position()]

            return command_list

    def move_below_slot(self, slot: int)->list[OverviewStatus]:
        command_list: list[OverviewStatus]
        command_list = [
            lambda: self.rotate_handler_to_slot(slot),
            lambda: self.move_x_to_slot(slot),
            lambda: self.move_handler_below_slot_height(slot)]
        return command_list

    def pipet_station(self, rows: int, distance_mm: float, start_mm: float) -> list[OverviewStatus]:
        if (rows * distance_mm) + start_mm > CS.steps_to_mm_shovel(self.max_steps_shovel):
            raise ValueError("The parameters are ot compatible with the Plate handler")
            
        command_list: list[OverviewStatus]
        command_list = [
            lambda: self.move_handler_below_slot_height(self.pipet_station_slot),
            lambda: self.run_shovel_in_relative_mm(start_mm)        
        ]
        for i in range(rows):
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_shovel_in_relative_mm(distance_mm))
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_height_in_relative_mm(15))
            command_list.append(lambda: time.sleep(0.5))
            command_list.append(lambda: self.run_height_in_relative_mm(-15))
            command_list.append(lambda: time.sleep(0.5))
            
        command_list.append(lambda: self.retract_shovel())
        return command_list

    def do_media_change_from_slot(self, slot: int):
        command_list: list[OverviewStatus]
        first_part = [  lambda: self.move_plate_from_slot_to_handler(slot),
                        lambda: self.leave_lid_at_holder()]

        second_part = self.move_below_slot(slot)

        third_part = self.pipet_station()

        fourth_part = [ lambda: self.get_lid_from_holder(),
                        lambda: self.move_plate_from_handler_to_slot(slot)]

        command_list = first_part + second_part + third_part + fourth_part
        return command_list


    #TODO finisch this func
    def inverted_do_mediachange_from_slot(self, slot: int):
        command_list: list[OverviewStatus]
        command_list = [
            lambda: self.move_plate_from_handler_to_slot(slot)]

    def get_plate_do_measurement_bring_back_plate(self, plate: str, slot_a: int) -> list[OverviewStatus]:
        command_list :list[OverviewStatus]
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(self.measurement_slot),
            lambda: self.run_foc(plate),
            lambda: self.move_plate_from_slot_to_handler(self.measurement_slot),
            lambda: self.move_plate_from_handler_to_slot(slot_a)]

        return command_list

    def inverted_get_plate_do_measurement_bring_back_plate(self, slot_a: int)->list[OverviewStatus]:
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(self.measurement_slot),
            lambda: time.sleep(0.1),
            lambda: self.move_plate_from_slot_to_handler(self.measurement_slot),
            lambda: self.move_plate_from_handler_to_slot(slot_a)]
        return command_list

    def move_plate_from_transfer_station_to_slot_v2(self):
        pass

    def inverted_move_plate_from_transfer_station_to_slot_v2(self):
        pass

    def move_plate_from_slot_a_to_slot_b(self, slot_a: int, slot_b:int):
        command_list :list[OverviewStatus]
        command_list = [
            lambda: self.move_plate_from_slot_to_handler(slot_a),
            lambda: self.move_plate_from_handler_to_slot(slot_b)]

        return command_list

    def run_foc(self, plate, _timeout: int =180):
        cmd_l = f'C:\\labhub\\Import\\FOC48.bat {plate}'
        
        try:
            process = subprocess.run(cmd_l, timeout=_timeout, shell = True)
            print("done")

        except subprocess.TimeoutExpired:
            print(" timeout expired")


    def test(self) -> list[OverviewStatus]:
        command_list: list[OverviewStatus]
        command_list = [
            lambda: self.open_transfer_door(),
            lambda: self.close_transfer_door()]
        return command_list

    def test3(self) -> list[OverviewStatus]:
        command_list: list[OverviewStatus]
        command_list = [
            lambda: print(1),
            lambda: print(2),
            lambda: print(3),
            lambda: print(4),
            lambda: print(5),
            lambda: print(6)]

        return command_list

    def inverted_test3(self):
        command_list = [
            lambda: print(6),
            lambda: print(5),
            lambda: print(4),
            lambda: print(3),
            lambda: print(2),
            lambda: print(1)]

        return command_list



import sqlite3

class CreateDB():
    def __init__(self, path):
        self.connect = sqlite3.connect(path)

        self.create_table_plate_actions()
        self.create_table_plates()
        self.create_table_slots()
        self.create_index_plate_id()
        self.create_index_assigned_plate_plate_id()
        self.create_index_current_plate_id()

        self.connect.close()

    def create_table_plate_actions(self):
        try:
            self.connect.execute(""" CREATE TABLE "PlateActions" (
                                    "Timestamp" TEXT NOT NULL CONSTRAINT "PK_PlateActions" PRIMARY KEY,
                                    "PlateId" TEXT NULL,
                                    "Action" TEXT NULL,
                                     CONSTRAINT "FK_PlateActions_Plates_PlateId" FOREIGN KEY ("PlateId") REFERENCES "Plates" ("PlateId") ON DELETE RESTRICT)""")
        except sqlite3.OperationalError:
            pass
    
    def create_table_plates(self):
        try:
            self.connect.execute(""" CREATE TABLE "Plates" (
                                    "PlateId" TEXT NOT NULL CONSTRAINT "PK_Plates" PRIMARY KEY,
                                    "Owner" TEXT NULL,
                                    "Location" INTEGER NOT NULL)""")
        except sqlite3.OperationalError:
            pass

    def create_table_slots(self):
        try:
            self.connect.execute(""" CREATE TABLE "Slots" (
                                    "SlotNumber" INTEGER NOT NULL CONSTRAINT "PK_Slots" PRIMARY KEY AUTOINCREMENT,
                                    "Role" INTEGER NOT NULL,
                                    "Occupied" INTEGER NOT NULL,
                                    "AssignedPlatePlateId" TEXT NULL,
                                    "CurrentPlateId" TEXT NULL,
                                    CONSTRAINT "FK_Slots_Plates_AssignedPlatePlateId" FOREIGN KEY ("AssignedPlatePlateId") REFERENCES "Plates" ("PlateId") ON DELETE RESTRICT,
                                    CONSTRAINT "FK_Slots_Plates_CurrentPlateId" FOREIGN KEY ("CurrentPlateId") REFERENCES "Plates" ("PlateId") ON DELETE RESTRICT) """)
        except sqlite3.OperationalError:
            pass

    def create_index_plate_id(self):
        try:
            self.connect.execute("""CREATE INDEX "IX_PlateActions_PlateId" ON "PlateActions" ("PlateId")""")
        except sqlite3.OperationalError:
            pass

    def create_index_assigned_plate_plate_id(self):
        try:
            self.connect.execute("""CREATE UNIQUE INDEX "IX_Slots_AssignedPlatePlateId" ON "Slots" ("AssignedPlatePlateId")""")
        except sqlite3.OperationalError:
            pass

    def create_index_current_plate_id(self):
        try:
            self.connect.execute("""CREATE INDEX "IX_Slots_CurrentPlateId" ON "Slots" ("CurrentPlateId")""")
        except sqlite3.OperationalError:
            pass

def create_db(path):
    CreateDB(path)


import shutil
import os
import sys
from pathlib import Path
from .create_db import create_db

def get_config_dir():
    return Path('C:/ProgramData/Cytomat')

def setupt_config_dir():
    config_dir = get_config_dir()
    db_file = config_dir / 'slots.db'

    try: 
        if not config_dir.exists():
            os.mkdir(config_dir)
            print(f"Created: {config_dir}")
        else:
            print(f" allready exits: {config_dir}")

    except Exception as e:
        print(f"""  Path directory could not be created: {e}
                    -
                    Please Create manualy: {config_dir}""")
    try:
        if not db_file.exists():
            create_db(db_file)
            print(f"copied sample DataBase to: {db_file}")
        else:
            print(f"DataBase file allready exits: {db_file}")

    except Exception as e:
        print(f"""  Error: {e}
                    -
                    Please run funktion create_db(Path) in the program cytomat/scripts/create_db.py
                    -
                    choose the Path: 'C:/ProgrammData/Cytomat/slots.db'""")
        
def post_install():
    setupt_config_dir()

if __name__ == '__main__':
    post_install()


from tkinter import *
from cytomat.utils import lazy_load_config_file
from cytomat.scripts.setup_cytomat import get_config_dir
from cytomat_plus.db import DB
import json
import re

class SettingManager():
    def __init__(self):
        self.__root = Tk()
        self.__root.title('Setting Manager')
        self.__root.geometry('1920x1080')
 
        self.config_data_json: dict = lazy_load_config_file()
        self.__total_columns_left = 3
    
    def get_root(self):
        return self.__root
    
    def setup_frames(self):
        self.__left_frame = Frame(self.__root)
        self.__left_frame.grid(row = 0, column = 0, sticky = 'nsew')
        self.__root.columnconfigure(0, weight=1)
                
        self.__right_frame = Frame(self.__root, bg = 'lightblue')
        self.__right_frame.grid(row = 0, column = 1, sticky = 'nsew')
        self.__root.columnconfigure(1, weight = 1)
        
        self.__root.rowconfigure(0, weight = 1)    

    def create_header_config(self):
        self.__header_config = Label(self.__left_frame, text = "Set specific parameters for the Cytomat", font=('Helvetica', 20), pady = 1, padx = 1)
        self.__header_config.grid(row = 0, column = 0, sticky = 'ns', columnspan = self.__total_columns_left)

    def create_left_frame_objects(self):
        self.__left_frame.grid_columnconfigure(0, weight=2)
        self.__left_frame.grid_columnconfigure(1, weight=2)
        self.__left_frame.grid_columnconfigure(2, weight=1)
        for index, (key, value) in enumerate(self.config_data_json.items()):    
            SettingsConfigFile(root = self.__left_frame, setting_manager = self, key = key, value = value, type = type(value), row_idx = index + 1, data=self.config_data_json)

    def create_right_frame_object(self):
        SettingsSlots(root = self.__right_frame)

    def create_overview(self):
        self.setup_frames()
        self.create_header_config()
        
        self.create_left_frame_objects()
        self.create_right_frame_object()

        #Create Rows to Setup Slots
        
    def submit_callback(self):
        pass

"""
This Class displays the left side of the window
"""

class SettingsConfigFile():
    __root: Tk
    def __init__(self, root, setting_manager, key, value, type, row_idx, data):
        self.setting_manager = setting_manager
        self.key = key
        self.value = value
        self.type = type
        self.__root = root
        self.row_idx = row_idx
        self.data = data
        self.create_row()

    def create_row(self):
        self.__root.rowconfigure(self.row_idx, weight =1)
        self.create_label()
        self.create_entry()
        self.create_enter_button()

    def create_label(self):
        self.__row_label = Label(self.__root, text = str(self.key), font=('Helvetica', 14), relief='solid',  pady = 5, padx = 1)
        self.__row_label.grid(row=self.row_idx, column = 0, sticky='nsew', ipadx = len(str(self.value)))

    def create_entry(self):
        entry_text = StringVar(value = self.value)
        self.__row_entry = Entry(self.__root, textvariable = entry_text)
        self.__row_entry.grid(row = self.row_idx, column = 1, sticky='nsew', ipadx = len(str(self.value)))

    def create_enter_button(self):
        self.enter_button = Button(self.__root, text = "Enter", command = self.enter_button_callback)
        self.enter_button.grid(row = self.row_idx, column = 2, sticky='nsew', ipadx = len('entry'))
        
    def enter_button_callback(self):
        entry = self.__row_entry.get()
        
        print(f"""  entry: {entry}
                    entry is digit: {entry.isdigit()}
                    row type: {self.type}""")
        
        if entry.isdigit() and self.type is not int:
            return
        
        if not entry.isdigit() and self.type is int:
            return
        
        if entry.isdigit() and self.type is int:
            entry = int(entry)
            self.data[self.key] = entry
            
        if not entry.isdigit() and self.type is str:
            print("COM_port_entry")
            #regular expression if the entry is for the COMPort
            comp = r"^COM\d+$" 
            
            if bool(re.match(comp, entry)) is False:
                return print("invalid COMPort")
            
            self.data[self.key] = entry
            
            
            
class SettingsSlots():
    __root: Frame
    def __init__(self, root):
        self.__root = root
        self.__db = DB()
        self.__slots_roled = self.__db.get_roled_slots()
        self.__slots_unroled = self.__db.get_unroled_slots()
        if len(self.__slots_unroled) > 0:
            print(self.__slots_unroled)
            self.__entry_dict = self.create_entry_dict()
            self.aplay_changes_display()
        else: 
            self.__entry_dict = {}
        self.create()
        
    def create(self):
        self.set_grid_settings()
        self.create_header()
        self.create_entry_area()    
    
    def create_header(self):
        self.__header_slots = Label(self.__root, text = "Set the Slots wich should be used", font = ('Helvetica', 20), pady = 1, padx = 1)
        self.__header_slots.grid(row = 0, column=0, columnspan = 4, sticky='nsew')
        
    def create_entry_area(self):
        self.__entry_bottom = Entry(self.__root, textvariable = 'bottom')
        self.__entry_bottom.grid(row = 1, column = 0, sticky = 'nsew')
        
        self.__middle_label = Label(self.__root, text = '-', font=('Helvetica', 18))
        self.__middle_label.grid(row= 1, column = 1, sticky = 'nsew')
        
        self.__entry_upper = Entry(self.__root, textvariable = 'upper')
        self.__entry_upper.grid(row = 1, column = 2, sticky = 'nsew')
    
        self.__enter_button = Button(self.__root, text = 'Enter', command = self.enter_button_callback)
        self.__enter_button.grid(row = 1, column = 3, sticky = 'nsew')
        
    def set_grid_settings(self):
        self.__root.columnconfigure(0, weight = 2)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.columnconfigure(2, weight = 2)
        self.__root.columnconfigure(3, weight = 1)
    
    """
    """
        
    def initialze_entry_object(self, sequence: list, db: DB):
        obj = Entries(sequence = sequence, db = db)
        return obj
    
    def aplay_changes_db(self):
        for key in self.__entry_dict.keys():
            start, end = key
            for slot in range(start, end + 1): 
                self.__db.insert_new_row(slot = slot, plate = None, role = 0, occupied = 0, current_plate_id = None)

        self.__slots_unroled = self.__db.get_unroled_slots()
        
    def aplay_changes_display(self):
        for index, key in enumerate (self.__entry_dict.keys()):
            obj: Entries = self.__entry_dict[key][1]
            obj.display_entry(row = index + 2, root = self.__root)
    
    def enter_button_callback(self):
        is_valid = self.is_entry_valid(entry_bottom = self.__entry_bottom.get(), entry_upper = self.__entry_upper.get())
        
        if not is_valid:
            return
        
        kind = self.kind_of_entry()
        if kind == 'new':
            self.create_new_entry()
            
        if kind == 'merge':
            self.merge_entries()
            
        self.aplay_changes_db()
        self.aplay_changes_display()
        
        
        """
        This function iterates through all the slots already in use for regular plates and searches for contiguous sequences.
        For each found sequence, it creates an Entries object (which is initialized with the sequence) and saves it in a dictionary
        using the start and end of the sequence as the key. It returns that dictionary.
        """
    def create_entry_dict(self)->dict:
        entry_dict = {}
        curr_subsequence = []
        for i in range (1, len(self.__slots_unroled)):
            if  self.__slots_unroled[i] == self.__slots_unroled[i-1] + 1:
                curr_subsequence.append(i) 
            else:
                key = (curr_subsequence[0], curr_subsequence[-1])
                entry_dict[key] = (curr_subsequence, self.initialze_entry_object(sequence=curr_subsequence, db = self.__db))
                curr_subsequence = []
                
        #last entry        
        key = (curr_subsequence[0], curr_subsequence[-1])
        entry_dict[key] = (curr_subsequence, self.initialze_entry_object(sequence=curr_subsequence, db = self.__db))
        
        return entry_dict
    
    def is_entry_valid(self, entry_bottom, entry_upper)-> bool:
        
        #checks if both of the entrys are positive integers
        if entry_bottom.isdigit() == False or entry_upper.isdigit() == False:
            return False
        
        if int(entry_bottom) > int(entry_upper):
            return False
        
        self.__entry_list = self.create_entry_list(int(entry_bottom), int(entry_upper))
           
        #checks if any of the numbers tryning to set as a Slot is already roled.
        if self.is_intersection_roled(self.__entry_list):
            return False
            
        return True
    
    """
    return 'merge' if there is an intersection or a countinuos sequence in the new entry and the used slots in db, else 'new' for new entry obj
    """
    def kind_of_entry(self)->str:
        if self.is_intersection_used(self.__entry_list):
            return 'merge'
        else: 
            return 'new'
        
    """Return a boolean indicating whether there is at least one element with a non-zero role that is included
       in the list of slots being attempted to set.
    """
        
    def is_intersection_roled(self, entry_list: list)->bool:
        return bool(set(entry_list) & set(self.__slots_roled))
    
    def is_intersection_used(self, entry_list: list)->bool:
        if not self.__slots_unroled:
            return False
        
        if bool(set(entry_list) & set(self.__slots_unroled)):
            return True
        
        min_entry, max_entry = min(entry_list), max(entry_list)
        min_slot, max_slot = min(self.__slots_unroled), max(self.__slots_unroled)

        if max_entry + 1 == min_slot or max_slot + 1 == min_entry:
            return True
        
        return False
        
    def create_entry_list(self, entry_bottom: int, entry_upper: int):
        entry_list = []
        for i in range(entry_bottom, entry_upper+1):
            entry_list.append(i)
        return entry_list
    
    """
    overlapping entries like 1-4 and 2-5 have to be merged. 
    """
    def merge_entries(self):
        start, end = self.__entry_list[0], self.__entry_list[-1]
        keys_to_remove = []
        
        for key in list(self.__entry_dict.keys()):
            seq_start, seq_end = key
            
            #checks if new sequence is identical to an old one or an other sequence contains the new one
            if start >= seq_start and end <= seq_end:
                return None
            
            #checks if new sequence is contains an other sequence 
            if start <= seq_start and end >= seq_end:
                print("hey2")
                keys_to_remove.append(key)

        for key in keys_to_remove:
            del self.__entry_dict[key] 

        #getting all overlapping sequences
        overlapping_keys = []
        for key in list(self.__entry_dict.keys()):
            seq_start, seq_end = key
            if (end >= seq_start or start <= seq_end):
                overlapping_keys.append(key)
        
        merged_list = self.__entry_list.copy()
        for key in overlapping_keys:
            merged_list += self.__entry_dict[key][0]
            del self.__entry_dict[key]
            
        merged_list = sorted(set(merged_list))
        new_key = (merged_list[0], merged_list[-1])
        
        self.__entry_dict[new_key] = (merged_list, self.initialze_entry_object(sequence = merged_list.copy(), db = self.__db))
        self.__slots_unroled = sorted(set(self.__slots_unroled + merged_list))
        
    def create_new_entry(self):
        start, end = self.__entry_list[0], self.__entry_list[-1]
        
        new_key = (start, end)
        self.__entry_dict[new_key] = (self.__entry_list, self.initialze_entry_object(sequence = self.__entry_list(), db=self.__db))
        
        self.__slots_unroled += sorted(set(self.__slots_unroled + self.__entry_list.copy()))
        
   
    def check_for_current_slots():
        pass        
    
    def update_overview(self):
        pass
    
class Entries():
    __DB: DB
    def __init__(self, sequence, db):
        self.sequence = sequence
        self.__start = sequence[0]
        self.__end = sequence[-1]
        self.__db = db

    def display_entry(self, row, root: Frame):
        row = row +2
        root.columnconfigure(0, weight = 2)
        root.columnconfigure(1, weight = 1)
        root.rowconfigure(row, weight = 1)
        self.create_label(row = row, root = root)
        self.create_delete_button(row = row, root = root)
        
    def create_label(self, row, root: Frame):
        self.__label = Label(root, text = f"{self.__start} - {self.__end}", font = ('Helvetica', 20), relief='solid')
        self.__label.grid(row = row, column = 0, sticky= 'nsew')
        
    def create_delete_button(self, row, root: Frame):
        self.__delete_Button = Button(root, text = "Delete Entry", command = self.delete_button_callback)
        self.__delete_Button.grid(row = row, column = 1, sticky='nsew')
        
    def delete_button_callback(self):
        print("Delete")

manager = SettingManager()
manager.create_overview()
root = manager.get_root()
root.mainloop()


from cytomat.cytomat import Cytomat
from cytomat.status import OverviewStatus
from .plate_handler_plus import PlateHandlerPlus
from .db import DB
from typing import Union
#from ws_barcode_scanner import BarcodeScanner
import time


#current firmware settings: left stack 1-21, right stack 22-42
class CytomatPlus(Cytomat):
    plate_handler: PlateHandlerPlus
    
    def __init__(self, serial_port: str):
        super().__init__(serial_port)
        self.plate_handler = PlateHandlerPlus(self._Cytomat__serial_port)
        self.db = DB()
        self.execute_error = None
        print("cytomat")
        #self.scanner = BarcodeScanner("COM3")

    def wait_until_not_busy(self, timeout: float  = 30, poll_intervall: float = 0.05):
        status = self.overview_status
        max_time = timeout
        start_time = time.time()
        while status.busy:
            if (time.time()-start_time) >= max_time:
                raise TimeoutError(f"Device still busy after {max_time} seconds")
            time.sleep(poll_intervall)
            status = self.overview_status

    def execute(self, command_list: Union[list[OverviewStatus], OverviewStatus], timeout: float = 30, poll_interval: float = 0.5)-> OverviewStatus:
        if not isinstance(command_list, list):
            command_list = [command_list]

        for command in command_list:
            self.wait_until_not_busy(timeout, poll_interval)
            try:
                command()
                self.wait_until_not_busy(timeout = timeout, poll_intervall = poll_interval)
            except Exception as e:
                self.plate_handler.initialize()

    def read_barcode(self):
        self.scanner.query_for_codes()

    def get_status(self):
        return self.overview_status


from tkinter import *
from cytomat.utils import lazy_load_config_file
from cytomat.scripts.setup_cytomat import get_config_dir
from .manage_settings.manager_slots import ManagerSlots
from .manage_settings.manager_config_file import ManagerConfigFile
from sqlite3 import IntegrityError
import tkinter.font

class ManagerSettings():
    __root: Toplevel
    def __init__(self, db, root): 
        self.__root = Toplevel(root)
        self.__root.title('Setting Manager')
        self.__root.geometry('1280x720')
        self.__root.withdraw()
        self.__db = db
        
        #from utils file
        self.__config_data_json: dict = lazy_load_config_file()
        #from setup_cytomat file
        self.__config_path = get_config_dir()
        
        self.sync_db_with_config_file()
        self.__root.protocol("WM_DELETE_WINDOW", self.on_close)
    
    def get_root(self):
        return self.__root

    def display_window(self):
        self.__root.deiconify()
    
    def create_overview(self):
        self.set_grid_settings()
        self.create_enter_label()
        self.create_button_gui()
        self.create_button_settings()
        self.create_theme_check_box()
        self.__theme_pack = self.set_theme('light')
        self.aplay_theme_settings(self.__theme_pack)
        
    def set_grid_settings(self):
        self.__root.rowconfigure(0, weight = 1)
        self.__root.rowconfigure(1, weight = 2)
        self.__root.columnconfigure(0, weight = 1)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.columnconfigure(2, weight = 1)
        
    def create_enter_label(self):
        font = tkinter.font.Font(family = 'Times New Roman', size = 14)
        
        self.__enter_label = Label(self.__root, text = 'Enter Settings', font=font)
        self.__enter_label.grid(row=0, column=0, sticky = 'ew')
        
        self.__enter_label_gui = Label(self.__root, text = 'Enter Cytomat GUI', font = font)
        self.__enter_label_gui.grid(row = 0, column=1, sticky='ew')
        
    def create_button_settings(self):
        self.__button_settings = Button(self.__root, text = 'Enter', command = self.settings_button_callback)
        self.__button_settings.grid(row = 1, column=0, ipadx = 40, ipady = 10, sticky='n')
        
    def create_button_gui(self):
        self.__button_gui = Button(self.__root, text = 'Enter', command = self.gui_button_callback)
        self.__button_gui.grid(row = 1, column=1, ipadx = 40, ipady = 10, sticky='n')
        
    def settings_button_callback(self):
        self.create_top_level_window()
    
    def gui_button_callback(self):
        self.__root.withdraw()
        
    def on_close(self):
        #self.no_button_callback()
        self.__root.destroy()
        
    def create_theme_check_box(self):
        self.check_box_var = BooleanVar()
        self.__theme_check_box = Checkbutton(self.__root, text = 'Dark Mode', variable= self.check_box_var, command = self.theme_check_box_callback)
        self.__theme_check_box.grid(row = 0, column=2)
        
    def theme_check_box_callback(self):
        if self.check_box_var.get():
            self.__theme_pack = self.set_theme('dark')
        else:
            self.__theme_pack = self.set_theme('light')
            
        self.aplay_theme_settings(self.__theme_pack)
    
    def set_theme(self, theme):
        theme_pack = {}
        if theme == 'dark':
            theme_pack['family'] = 'dark'
            theme_pack['entry'] = '#616161'
            theme_pack['button'] = '#424242'
            theme_pack['fg'] = 'white'
            theme_pack['bg'] = '#454545'
            theme_pack['header'] = '#707070'
            
        else:
            theme_pack['family'] = 'light'
            theme_pack['entry'] = 'lightgrey'
            theme_pack['button'] = 'white'
            theme_pack['fg'] = 'black'
            theme_pack['bg'] = '#E5E5E5'
            theme_pack['header'] = '#BDBDBD'
            
        return theme_pack
    
    def aplay_theme_settings(self, theme_pack: dict):
        bg = theme_pack['bg']
        fg = theme_pack['fg']
        entry = theme_pack['entry']
        button = theme_pack['button']
        
        self.__root.config(bg=bg)
        self.__button_settings.config(bg = button, fg = fg)
        self.__button_gui.config(bg = button, fg = fg)
        self.__enter_label.config(fg = fg, bg = bg)
        self.__enter_label_gui.config(bg = bg, fg = fg)
        
        try:
            self.top_level.applay_theme_pack(theme_pack=self.__theme_pack)
            
        except Exception:
            pass
            
    def create_top_level_window(self):
        self.top_level = ManagerSettingsTop(root = self.__root, config_data=self.__config_data_json, config_path=self.__config_path, db = self.__db)
        self.top_level.create_overview()
        self.top_level.applay_theme_pack(theme_pack=self.__theme_pack)
    
    def sync_db_with_config_file(self):
        for key, value in self.__config_data_json.items():
            if not 'slot' in key:
                continue
            
            if not value:
                continue
            
            try:    
                self.__db.insert_new_row(slot = value, role = 1)
                 
            except IntegrityError:
                continue


        #Create Rows to Setup Slots
        
    def submit_callback(self):
        pass
    
    
class ManagerSettingsTop():
    __root: Toplevel
    def __init__(self, root, db, config_path, config_data):
        self.__root = Toplevel(root)
        self.__root.title('Settings')
        self.__root.geometry('1920x1080')
        
        self.__db = db
        self.__config_data_json = config_data
        self.__config_path = config_path
        self.setup_frames()
        
    def get_root(self):
        return self.__root
        
    def create_overview(self):
        self.__manager_config = ManagerConfigFile(root = self.__left_frame, data = self.__config_data_json, config_directory_path = self.__config_path, db=self.__db)
        self.__manager_slots = ManagerSlots(root = self.__right_frame, db = self.__db)
        
        self.__manager_config.create_overview()
        self.__manager_slots.create_overview()
        
        
    def setup_frames(self):
        self.__left_frame = Frame(self.__root)
        self.__left_frame.grid(row = 0, column = 0, sticky = 'nsew')
                
        self.__right_frame = Frame(self.__root, bg = 'lightblue')
        self.__right_frame.grid(row = 0, column = 1, sticky = 'nsew')
        
        self.__root.columnconfigure(0, weight=1)
        self.__root.columnconfigure(1, weight = 1)
        self.__root.rowconfigure(0, weight = 1)
    
    def applay_theme_pack(self, theme_pack):
        self.__theme_pack = theme_pack
        self.__manager_config.aplay_theme_pack(theme_pack = self.__theme_pack)
        self.__manager_slots.aplay_theme_pack(theme_pack = self.__theme_pack)